"""
Application for configuring and building the app environments.
"""

import os, re, sys
import logging
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urlparse

from traitlets import Unicode, Integer, Bool, Dict, validate, Any, default, observe, List, TraitError
from traitlets.config import Application, catch_config_error, SingletonConfigurable
from tornado.httpclient import AsyncHTTPClient
from tornado.httpserver import HTTPServer
import tornado.ioloop
import tornado.options
import tornado.log
from sqlalchemy.exc import OperationalError
from jinja2 import Environment, FileSystemLoader, PrefixLoader, ChoiceLoader
from jupyterhub.services.auth import HubOAuthCallbackHandler
from jupyterhub import __version__ as __jh_version__
from jupyterhub import dbutil
from jupyterhub.traitlets import EntryPointType

from .dashboard import DashboardRepr
from .util import url_path_join
from jupyterhub import orm as jhorm
from .builder.builders import BuildersDict, Builder
from ._data import DATA_FILES_PATH
from .pluggymanager import pm
from . import hookimpl


_TEMPLATE_PATH_OPEN = os.path.join(DATA_FILES_PATH, 'templates-open')
_TEMPLATE_PATH_RESTRICTED = os.path.join(DATA_FILES_PATH, 'templates-restricted')
_TEMPLATE_PATH_COMMON = os.path.join(DATA_FILES_PATH, 'templates-common')


CDS_TEMPLATE_PATHS = [_TEMPLATE_PATH_OPEN, _TEMPLATE_PATH_COMMON]
CDS_TEMPLATE_PATHS_RESTRICTED = [_TEMPLATE_PATH_RESTRICTED, _TEMPLATE_PATH_COMMON]

_all_allowed_presentation_types = ['voila', 'streamlit', 'plotlydash', 'bokeh', 'rshiny']


class CDSDashboardsConfig(SingletonConfigurable):

    builder_class = EntryPointType(
        default_value='cdsdashboards.builder.builders.Builder',
        klass=Builder,
        entry_point_group="cdsdashboards.builders",
        help="""The class to use for building dashboard servers.

        Should be a subclass of :class:`cdsdashboards.builder.builders.Builder`.

        May be registered via entry points,
            e.g. `c.cdsdashboards.builders = 'localprocess'`
        """,
    ).tag(config=True)

    server_name_template = Unicode(
        'dash-{urlname}',
        help="""
        How to name the final user server that runs the dashboard. Template vars will be expanded:

        {urlname} : dashboard URL-safe name
        {date} : <current date in YYmmdd format>
        {time} : <current date in HHMMSS format>

        """
    ).tag(config=True)

    presentation_types = List(
        trait=Unicode,
        default_value=_all_allowed_presentation_types,
        minlen=1,
        help="""
        Allowed presentation types for Dashboards. A list, allowed strings are: {}.
        There must be at least one valid entry.
        Default value is all the allowed presentation types.
        Add any custom frameworks to the extra_presentation_types config if you want to augment instead of overwrite the default list.
        """.format(_all_allowed_presentation_types)
    ).tag(config=True)

    extra_presentation_types = List(
        trait=Unicode,
        default_value=[],
        minlen=0,
        help="""
        Extra custom presentation types for Dashboards, to be added to the presentation_types list.
        A list.
        Default value is the empty list.
        """
    ).tag(config=True)

    @property
    def merged_presentation_types(self):
        return self.presentation_types + self.extra_presentation_types

    show_source_servers = Bool(
        False,
        help="""
        Allow the user to select a source server when creating a Dashboard (currently only relevant for DockerSpawner).
        """
    ).tag(config=True)

    require_source_server = Bool(
        False,
        help="""
        Require the user to select a source server when creating a Dashboard (currently only relevant for DockerSpawner).
        You must set show_source_servers to True if you set require_source_server to True.
        """
    ).tag(config=True)

    show_source_git = Bool(
        True,
        help="""
        Allow the user to enter a git repo to fetch files for a dashboard. 
        """
    ).tag(config=True)

    default_allow_all = Bool(
        True,
        help="""
        If True (default) then newly-created Dashboards will be accessible to all authenticated JupyterHub users.
        If False, only members of the dashboard's own group will be allowed to access it.
        This flag is passed on to the Dashboard object's allow_all field so can be subsequently overridden in the database.
        """
    ).tag(config=True)

    conda_envs = List(
        trait=Unicode,
        default_value=[],
        minlen=0,
        help="""
        A list of Conda env names for the dashboard creator to select.
        A list.
        Default value is the empty list.
        """
    ).tag(config=True)

    allow_custom_conda_env = Bool(
        False,
        help="""
        If True then dashboard creators can type any value for the Conda env to use, even if none is provided in the conda_envs list of pre-defined env names.
        If False (default), only Conda env names listed in the conda_envs setting can be selected - if any are present.
        """
    ).tag(config=True)

    spawn_allow_group = Unicode(
        '',
        help="""
        Name of a JupyterHub group whose users should be allowed to spawn servers and create dashboards.
        See also spawn_block_group.
        """
    ).tag(config=True)

    spawn_block_group = Unicode(
        '',
        help="""
        Name of a JupyterHub group whose users should be blocked from spawning servers and creating dashboards.
        If blank, spawn_allow_group will determine which users should be able to spawn.
        If both settings are blank, all users will be allowed to spawn.
        If both are non-blank, spawn_block_group will take priority.
        """
    ).tag(config=True)
    

class CDSConfigStore():

    _instance = None

    @classmethod
    def get_instance(cls, config):
        """
        Supply a config object to get the singleton CDSDashboardsConfig instance - only normally available from web handlers
        """
        if cls._instance:
            return cls._instance
        
        cls._instance = CDSDashboardsConfig(config=config)

        return cls._instance


class BuildersStore():

    _instance = None

    @classmethod
    def get_instance(cls, config):
        """
        Supply a config object to get the singleton instance - only normally available from web handlers
        """
        if cls._instance:
            return cls._instance
        
        cdsconfig = CDSConfigStore.get_instance(config)

        builder_class = cdsconfig.builder_class

        def builder_factory(dashboard):
            return builder_class(dashboard=dashboard, cdsconfig=cdsconfig)

        cls._instance = BuildersDict(builder_factory)
        return cls._instance
        

cds_tornado_settings = {}



import sys
from concurrent.futures import ThreadPoolExecutor
import functools

import docker
from docker.utils import kwargs_from_env
from traitlets import Dict, Unicode, Any
from tornado import ioloop
from tornado.log import app_log
from .builders import BuildException
from .processbuilder import ProcessBuilder
from .. import hookimpl
from ..pluggymanager import pm


class BasicDockerBuilder(ProcessBuilder):

    user = Any()

    @property
    def client(self):
        """single global client instance"""
        cls = self.__class__
        if cls._client is None:
            kwargs = {"version": "auto"}
            if self.tls_config:
                kwargs["tls"] = docker.tls.TLSConfig(**self.tls_config)
            kwargs.update(kwargs_from_env())
            kwargs.update(self.client_kwargs)
            client = docker.APIClient(**kwargs)
            cls._client = client
        return cls._client

    tls_config = Dict(
        config=True,
        help="""Arguments to pass to docker TLS configuration.
        See docker.client.TLSConfig constructor for options.
        """,
    )

    client_kwargs = Dict(
        config=True,
        help="Extra keyword arguments to pass to the docker.Client constructor.",
    )

    _client = None

    def _docker(self, method, *args, **kwargs):
        """wrapper for calling docker methods
        to be passed to ThreadPoolExecutor
        """
        m = getattr(self.client, method)
        return m(*args, **kwargs)

    def docker(self, method, *args, **kwargs):
        """Call a docker method in a background thread
        returns a Future
        """
        fn = functools.partial(self._docker, method, *args, **kwargs)
        return ioloop.IOLoop.current().run_in_executor(self._executor, fn)

    _executor = None

    @property
    def executor(self):
        """single global executor"""
        cls = self.__class__
        if cls._executor is None:
            cls._executor = ThreadPoolExecutor(1)
        return cls._executor

    repo_prefix = Unicode(default_value='cdsuser').tag(config=True)

    async def prespawn_server_options(self, dashboard, dashboard_user, ns):

        source_spawner_orm = dashboard.source_spawner

        if source_spawner_orm is None:
            return {} # Just use default image

        tag = self.format_string('{date}-{time}', ns=ns)

        source_spawner_name = source_spawner_orm.name

        #if source_spawner_name not in dashboard_user.spawners:
        #    raise BuildException('The source server is not currently available for this dashboard')

        source_spawner = dashboard_user.spawners[source_spawner_name]

        app_log.debug('source_spawner {}'.format(source_spawner))

        #if source_spawner.state is None:
        #    raise BuildException('Source server has never been run, so there is nothing to clone!')

        object_id = source_spawner.object_id #source_spawner.state.get('object_id', None)

        app_log.debug('Docker object_id is {}'.format(object_id))

        if object_id is None:
            raise BuildException('No docker object specified in spawner state - maybe the source server has never been run')

        source_container = await self.docker('inspect_container', object_id)

        if source_container is None:
            raise BuildException('No docker object returned as source container')

        # Commit image of current server

        reponame = '{}/{}'.format(self.repo_prefix, dashboard.urlname)

        image_name = '{}:{}'.format(reponame, tag)

        app_log.info('Committing Docker image {}'.format(image_name))

        dockerfile_changes="\n".join([
            'CMD ["{}-entrypoint.sh"]'.format(dashboard.presentation_type),
            'ENV JUPYTERHUB_GROUP {}'.format(dashboard.groupname),
            'ENV JUPYTERHUB_ANYONE {}'.format(dashboard.allow_all and '1' or '0'),
            'ENV JUPYTERHUB_CDS_PRESENTATION_PATH "{}"'.format(dashboard.start_path or ''),
        ])

        await self.docker('commit', object_id, repository=reponame, tag=tag, changes=dockerfile_changes)

        self.log.info('Finished commit of Docker image {}:{}'.format(reponame, tag))

        return {'image': image_name}


DockerBuilder = BasicDockerBuilder
# Register plugin hooks so we use the Basic handlers by default, unless overridden

@hookimpl
def get_builder_DockerBuilder():
    return BasicDockerBuilder

pm.register(sys.modules[__name__])

DockerBuilder = pm.hook.get_builder_DockerBuilder()[0]



from asyncio import sleep
from datetime import datetime

from traitlets.config import LoggingConfigurable
from traitlets import Any, Bool, Integer
from tornado.log import app_log


class Builder(LoggingConfigurable):
    """Base class for building a dashboard, e.g. by cloning an existing server which is a Docker container.

    Subclass this, and override the following methods:

    - start

    """

    dashboard = None

    cdsconfig = None

    # private attributes for tracking status
    _build_pending = False
    _build_future = None

    event_queue = []

    @property
    def _failed(self):
        """Did the last build fail?"""
        return (
            not self.active
            and self._build_future
            and self._build_future.done()
            and self._build_future.exception()
        )

    @property
    def _log_name(self):
        """Return username:dashboard_urlname
        """
        if self.dashboard:
            return '%s:%s' % (self.dashboard.user.name, self.dashboard.urlname)
        else:
            return 'Dashboard Builder {}'.format(self)

    @property
    def pending(self):
        """Return the current pending event, if any

        Return False if nothing is pending.
        """
        if self._build_pending:
            return 'build'

        return None

    @property
    def ready(self):
        """Is this builder finished, up to assigning a spawner (which may still be starting up)

        A builder is not ready if an event is pending.
        """
        if self.pending:
            return False
        if self.dashboard is None:
            return False
        if self.dashboard.final_spawner is None:
            return False
        if self._build_future and self._build_future.done() and self._build_future.exception():
            return False
        return True

    @property
    def active(self):
        """Return True if the server is active.

        This includes fully running and ready or any pending start/stop event.
        """
        return bool(self.pending or self.ready)

    # options passed by constructor
    orm_builder = Any()
    db = Any()

    log = Any(default_value=app_log).tag(config=True)

    user = Any()

    def __init_subclass__(cls, **kwargs):

        app_log.debug('init subclass of builder')

        super().__init_subclass__()

        missing = []
        for attr in ('start',):
            if getattr(Builder, attr) is getattr(cls, attr):
                missing.append(attr)

        if missing:
            raise NotImplementedError(
                "class `{}` needs to redefine the `start` method(s). `{}` not redefined.".format(
                    cls.__name__, '`, `'.join(missing)
                )
            )

    def __init__(self, dashboard, cdsconfig, config=None):
        super().__init__(config=config)

        self.dashboard = dashboard
        self.cdsconfig = cdsconfig

    @property
    def last_activity(self):
        return self.orm_builder.last_activity

    consecutive_failure_limit = Integer(
        0,
        help="""
        Maximum number of consecutive failures to allow before
        shutting down JupyterHub.

        This helps JupyterHub recover from a certain class of problem preventing launch
        in contexts where the Hub is automatically restarted (e.g. systemd, docker, kubernetes).

        A limit of 0 means no limit and consecutive failures will not be tracked.
        """,
    ).tag(config=True)

    start_timeout = Integer(
        60,
        help="""
        Timeout (in seconds) before giving up on starting of single-user server.

        This is the timeout for start to return, not the timeout for the server to respond.
        Callers of spawner.start will assume that startup has failed if it takes longer than this.
        start should return when the server process is started and its location is known.
        """,
    ).tag(config=True)

    debug = Bool(False, help="Enable debug-logging of the single-user server").tag(
        config=True
    )

    async def _generate_progress(self):
        """Private wrapper of progress generator

        This method is always an async generator and will always yield at least one event.
        """
        if not self._build_pending:
            self.log.warning(
                #"Build not pending, can't generate progress for %s", self._log_name
                "Build not pending, can't generate progress"
            )
            return

        yield {"progress": 0, "message": "Builder requested"}
        from async_generator import aclosing

        async with aclosing(self.progress()) as progress:
            async for event in progress:
                yield event

    async def progress(self):
        """Async generator for progress events

        Must be an async generator

        Should yield messages of the form:

        ::

          {
            "progress": 80, # integer, out of 100
            "message": text, # text message (will be escaped for HTML)
            "html_message": html_text, # optional html-formatted message (may have links)
          }

        In HTML contexts, html_message will be displayed instead of message if present.
        Progress will be updated if defined.
        To update messages without progress omit the progress field.

        """
        next_event = 0

        break_while_loop = False
        while True:
            # Ensure we always capture events following the start_future
            # signal has fired.
            if self._build_future.done():
                break_while_loop = True
            event_queue = self.event_queue

            len_events = len(event_queue)
            if next_event < len_events:

                for i in range(next_event, len_events):
                    event = event_queue[i]

                    yield event 
                
                next_event = len_events

            if break_while_loop:
                break

            await sleep(1)

    def add_progress_event(self, event):
        self.event_queue.append(event)

    async def start(self, dashboard, dashboard_user, db):
        """Start the dashboard

        Returns:
          (str, str): the (new_server_name, new_server_options) of the new dashboard server.

        """
        raise NotImplementedError(
            "You must specify a c.CDSDashboardsConfig.builder_class in your JupyterHub jupyterhub_config.py file."
        )

    allow_named_servers = True # TODO take from main app config

    def template_namespace(self):
        """Return the template namespace for format-string formatting.

        Subclasses may add items to the available namespace.

        The default implementation includes::

            {
              'urlname': dashboard.urlname,
              'date': <current date in YYmmdd format>,
              'time': <current date in HHMMSS format>,
            }

        Returns:

            ns (dict): namespace for string formatting.
        """
        date = datetime.today().strftime('%Y%m%d')
        time = datetime.today().strftime('%H%M%S')
        d = {
            'urlname': self.dashboard.urlname,
            'date': date,
            'time': time
            }
        return d

    def format_string(self, s, ns=None):
        """Render a Python format string

        Uses :meth:`Builder.template_namespace` to populate format namespace, based on self.dashboard.

        Optionally provide the namespace as ns.

        Args:

            s (str): Python format-string to be formatted.

        Returns:

            str: Formatted string, rendered
        """
        if ns is None:
            ns = self.template_namespace()
        return s.format(**ns)


class BuildersDict(dict):
    def __init__(self, builder_factory):
        self.builder_factory = builder_factory

    def __getitem__(self, orm_dashboard):
        if orm_dashboard not in self:
            self[orm_dashboard] = self.builder_factory(orm_dashboard)
        return super().__getitem__(orm_dashboard)


class BuildException(Exception):
    pass



from tornado import web
from tornado.log import app_log
#from jupyterhub.services.auth import HubOAuthenticated, HubOAuth


class BaseHandler(web.RequestHandler): # HubOAuthenticated
    """HubAuthenticated by default allows all successfully identified users (see allow_all property)."""

    def initialize(self):
        super().initialize()
        self.log = app_log
        #self.hub_auth = HubOAuth.instance(config=self.settings['traitlets_config'])

    @property
    def template_namespace(self):
        return dict(static_url=self.static_url,
                    **self.settings.get('template_variables', {}))

    def render_template(self, name, **extra_ns):
        """Render an HTML page"""
        ns = {}
        ns.update(self.template_namespace)
        ns.update(extra_ns)
        template = self.settings['jinja2_env'].get_template(name)
        html = template.render(**ns)
        self.write(html)




import asyncio

from tornado.web import authenticated, HTTPError
from async_generator import aclosing

from jupyterhub.apihandlers.users import SpawnProgressAPIHandler
from jupyterhub.utils import iterate_until

from ..orm import Dashboard
from ..util import url_path_join
from ..app import BuildersStore


class ProgressDashboardHandler(SpawnProgressAPIHandler):

    @authenticated
    async def get(self, dashboard_urlname=''):
        self.set_header('Cache-Control', 'no-cache')

        current_user = await self.get_current_user()

        dashboard = self.db.query(Dashboard).filter(Dashboard.urlname==dashboard_urlname).one_or_none()

        if dashboard is None:
            raise HTTPError(404, 'No such dashboard or user')

        if not dashboard.is_orm_user_allowed(current_user.orm_user):
            raise HTTPError(403, 'User {} not authorized to access dashboard {}'.format(current_user.name, dashboard.urlname))
        

        # start sending keepalive to avoid proxies closing the connection
        asyncio.ensure_future(self.keepalive())
        # cases:
        # - spawner already started and ready
        # - spawner not running at all
        # - spawner failed
        # - spawner pending start (what we expect)
        def ready_event(dashboard):
            url = url_path_join(self.settings['base_url'], "user", dashboard.user.name, dashboard.final_spawner.name)
            return {
                'progress': 100,
                'ready': True,
                'message': "Redirecting to server at {}".format(url),
                'html_message': 'Redirecting to server at <a href="{0}" target="_blank">{0}</a>'.format(url),
                'url': url,
            }

        failed_event = {'progress': 100, 'failed': True, 'message': "Build failed or unable to get progress", 
            'url': url_path_join(self.settings['base_url'], "hub", "dashboards", dashboard.urlname, 'clear-error')
            }


        builders_store = BuildersStore.get_instance(self.settings['config'])

        builder = builders_store[dashboard]

        if builder.ready:
            # spawner already ready. Trigger progress-completion immediately
            self.log.info("Server %s is already started", builder._log_name)
            await self.send_event(ready_event(dashboard))
            return

        build_future = builder._build_future

        if not builder._build_pending:
            # not pending, no progress to fetch
            # check if spawner has just failed
            f = build_future
            if f and f.done() and f.exception():
                failed_event['message'] = "Build failed: %s" % f.exception()
                await self.send_event(failed_event)
                return
            else:
                raise HTTPError(400, "%s is not starting...", builder._log_name)

        if build_future: # Just in case build_future is None so iterate_until's asyncio.wait doesn't fail; 
            # builder._generate_progress should return an iterator because _build_pending was non-None
            # just above, with no awaits since

            # Retrieve progress events from the Builder
            async with aclosing(
                iterate_until(build_future, builder._generate_progress())
            ) as events:
                try:
                    async for event in events:
                        # don't allow events to sneakily set the 'ready' flag
                        if 'ready' in event:
                            event.pop('ready', None)
                        await self.send_event(event)
                except asyncio.CancelledError:
                    pass

            # progress finished, wait for spawn to actually resolve,
            # in case progress finished early
            # (ignore errors, which will be logged elsewhere)
            await build_future

        # progress and spawn finished, check if spawn succeeded
        if builder.ready:
            # spawner is ready, signal completion and redirect
            self.log.info("Server %s is ready", builder._log_name)
            await self.send_event(ready_event(dashboard))
        else:
            # what happened? Maybe spawn failed?
            f = build_future
            if f and f.done() and f.exception():
                failed_event['message'] = "Build failed: %s" % f.exception()
            else:
                self.log.warning(
                    "Server %s didn't start for unknown reason", builder._log_name
                )
            await self.send_event(failed_event)






from systemdspawner import SystemdSpawner

from .variablemixin import VariableMixin, MetaVariableMixin


class VariableSystemdSpawner(SystemdSpawner, VariableMixin, metaclass=MetaVariableMixin):

    def _expand_user_vars(self, string):
        """
        Expand user related variables in a given string

        Currently expands:
          {USERNAME} -> Name of the user
          {USERID} -> UserID
          {DASHSERVERNAME} -> A Dash plus Server Name (or '' if server is My Server which has no name)
          Also needs to preserve --, port, and base_url to pass on to jhsingle-native-proxy, so don't use straight format function
        """
        dsn = ''
        if self.name:
            dsn = '-{}'.format(self.name)
        return string.replace('{USERNAME}', self.user.name).replace('{USERID}', str(self.user.id)).replace('{DASHSERVERNAME}', dsn)




from tornado.log import app_log

from .builders import Builder, BuildException


class ProcessBuilder(Builder):
 
    async def start(self, dashboard, dashboard_user, db):
        """Start the dashboard

        Returns:
          (str, str): the (new_server_name, new_server_options) of the new dashboard server.

        """

        app_log.info('Starting Builder start function')

        self.event_queue = []

        self.add_progress_event({'progress': 10, 'message': 'Starting builder'})

        self._build_pending = True

        ns = self.template_namespace()

        new_server_name = self.format_string(self.cdsconfig.server_name_template, ns=ns)

        new_server_options = await self.prespawn_server_options(dashboard, dashboard_user, ns)

        if not self.allow_named_servers:
            raise BuildException(400, "Named servers are not enabled.")

        spawner = dashboard_user.spawners[new_server_name] # Could be orm_spawner or Spawner wrapper

        if spawner.ready:
            # include notify, so that a server that died is noticed immediately
            # set _spawn_pending flag to prevent races while we wait
            spawner._spawn_pending = True
            try:
                state = await spawner.poll_and_notify()
            finally:
                spawner._spawn_pending = False

        git_repo = dashboard.options.get('git_repo', '')
        conda_env = dashboard.options.get('conda_env', '')

        new_server_options.update({
            'presentation_type': dashboard.presentation_type or 'voila',
            'presentation_path': dashboard.start_path,
            'git_repo': git_repo,
            'conda_env': conda_env})
        
        if 'environment' not in new_server_options:
            new_server_options['environment'] = {}

        new_server_options['environment'].update({
                'JUPYTERHUB_ANYONE': '{}'.format(dashboard.allow_all and '1' or '0'),
                'JUPYTERHUB_GROUP': '{}'.format(dashboard.groupname)
                })

        return (new_server_name, new_server_options)

    async def prespawn_server_options(self, dashboard, dashboard_user, ns):
        return {} # Empty options - override in subclasses if needed




import os

from traitlets import Bool
from traitlets.config import SingletonConfigurable

from jupyterhub.handlers.static import CacheControlStaticFilesHandler

from .main import AllDashboardsHandler, DashboardEditHandler, MainViewDashboardHandler, ClearErrorDashboardHandler, \
        UpgradeDashboardsHandler, GroupsAllHandler, GroupsSingleHandler
from .core import OurHomeHandler
from .events import ProgressDashboardHandler
from .._data import DATA_FILES_PATH
from .api import DashboardsAPIHandler, DashboardDeleteAPIHandler, UserSelfAPIHandler


cds_extra_handlers = [
    
    (r'dashboards-db-upgrade', UpgradeDashboardsHandler),

    (r'dashboards-new', DashboardEditHandler),

    (r'dashboards', AllDashboardsHandler),
    
    (r'dashboards/(?P<dashboard_urlname>[^/]+?)/edit', DashboardEditHandler, {}, 'cds_dashboard_config_handler'),
    (r'dashboards/(?P<dashboard_urlname>[^/]+?)', MainViewDashboardHandler),
    (r'dashboards/(?P<dashboard_urlname>[^/]+?)/clear-error', ClearErrorDashboardHandler),

    (r'dashboards-static/(.*)', CacheControlStaticFilesHandler, dict(path=os.path.join(DATA_FILES_PATH, 'static'))),

    (r'dashboards-api', DashboardsAPIHandler),
    (r'dashboards-api/hub-info/user', UserSelfAPIHandler),
    (r'dashboards-api/(?P<dashboard_urlname>[^/]+?)', DashboardDeleteAPIHandler),
    (r'dashboards-api/(?P<dashboard_urlname>[^/]+?)/progress', ProgressDashboardHandler),

    (r'groupslist', GroupsAllHandler),
    (r'groupslist/(?P<groupname>[^/]+?)', GroupsSingleHandler),

    (r'home-cds', OurHomeHandler),

]

def config_for_dashboards(c):
    """
    Add the required configuration to the Configurable object to enable Dashboards
    TODO extend any settings that are already present in c
    """

    from ..app import CDS_TEMPLATE_PATHS, cds_tornado_settings

    c.JupyterHub.extra_handlers = cds_extra_handlers
    c.JupyterHub.tornado_settings = cds_tornado_settings
    c.JupyterHub.template_paths = CDS_TEMPLATE_PATHS

    return c

__all__ = ['cds_extra_handlers', 'config_for_dashboards']



from traitlets import Dict, Unicode, List
from tljh import user
from tljh.normalize import generate_system_username

from .variablesystemd import VariableSystemdSpawner


class VariableUserCreatingSpawner(VariableSystemdSpawner):
    """
    VariableSystemdSpawner with user creation on spawn.

    Mirrors UserCreatingSpawner in TLJH.

    FIXME: Remove this somehow?
    """
    user_groups = Dict(key_trait=Unicode(), value_trait=List(Unicode()), config=True)

    def start(self):
        """
        Perform system user activities before starting server
        """
        # FIXME: Move this elsewhere? Into the Authenticator?
        system_username = generate_system_username('jupyter-' + self.user.name)

        # FIXME: This is a hack. Allow setting username directly instead
        self.username_template = system_username
        user.ensure_user(system_username)
        user.ensure_user_group(system_username, 'jupyterhub-users')
        if self.user.admin:
            user.ensure_user_group(system_username, 'jupyterhub-admins')
        else:
            user.remove_user_group(system_username, 'jupyterhub-admins')
        if self.user_groups:
            for group, users in self.user_groups.items():
                if self.user.name in users:
                    user.ensure_user_group(system_username, group)
        return super().start()
        


from dockerspawner import DockerSpawner, SwarmSpawner, SystemUserSpawner

from .variablemixin import VariableMixin, MetaVariableMixin, Command


class VariableDockerSpawner(DockerSpawner, VariableMixin, metaclass=MetaVariableMixin):
    
    default_presentation_cmd = Command(
        ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'],
        allow_none=False,
        help="""
        The command to run presentations through jhsingle_native_proxy, can be a string or list.
        Default is ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main']
        for the singleuser Docker images to ensure hooks can run.
        """
    ).tag(config=True)


class VariableSwarmSpawner(SwarmSpawner, VariableMixin, metaclass=MetaVariableMixin):
    
    default_presentation_cmd = Command(
        ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'],
        allow_none=False,
        help="""
        The command to run presentations through jhsingle_native_proxy, can be a string or list.
        Default is ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main']
        for the singleuser Docker images to ensure hooks can run.
        """
    ).tag(config=True)


class VariableSystemUserSpawner(SystemUserSpawner, VariableMixin, metaclass=MetaVariableMixin):
    
    default_presentation_cmd = Command(
        ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'],
        allow_none=False,
        help="""
        The command to run presentations through jhsingle_native_proxy, can be a string or list.
        Default is ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main']
        for the singleuser Docker images to ensure hooks can run.
        """
    ).tag(config=True)



import json

from tornado.web import authenticated, HTTPError

from jupyterhub.apihandlers.base import APIHandler
from jupyterhub.utils import url_path_join

from ..orm import Dashboard
from .base import DashboardBaseMixin
from ..app import BuildersStore


class DashboardBaseAPIHandler(APIHandler, DashboardBaseMixin):
    
    def check_referer(self):  # These request is allowed from user server
        """Check Origin for cross-site API requests.

        Copied from WebSocket with changes:

        - allow unspecified host/referer (e.g. scripts)
        """
        host = self.request.headers.get("Host")
        referer = self.request.headers.get("Referer")

        # If no header is provided, assume it comes from a script/curl.
        # We are only concerned with cross-site browser stuff here.
        if not host:
            self.log.warning("Blocking API request with no host")
            return False
        if not referer:
            self.log.warning("Blocking API request with no referer")
            return False

        host_path = url_path_join(host, self.hub.base_url)
        user_path = url_path_join(host, 'user/')
        referer_path = referer.split('://', 1)[-1] + '/'
        if not referer_path.startswith(host_path) and not referer_path.startswith(user_path):
            self.log.warning(
                "Blocking Cross Origin API request.  Referer: %s, Host: %s, User host: %s",
                referer,
                host_path,
                user_path
            )
            return False

        return True


class DashboardsAPIHandler(DashboardBaseAPIHandler):

    @authenticated
    async def get(self):
        """Return the list of dashboards for the current user."""

        current_user = await self.get_current_user()

        def to_json(dashboard):
            return {
                "name": dashboard.name,
                "url": "{}hub/dashboards/{}".format(self.settings['base_url'], dashboard.urlname),
                "description": dashboard.description,
                "path": dashboard.start_path,
                "username": dashboard.user.name
            }

        my_dashboards = current_user.dashboards_own
        body = {"_owned": list(map(to_json, my_dashboards))}

        visitor_dashboard_groups = self.get_visitor_dashboards(current_user)
        for username, dashboards in visitor_dashboard_groups.items():
            body[username] = list(map(to_json, dashboards))

        self.set_status(200)
        self.finish(json.dumps(body))


class DashboardDeleteAPIHandler(DashboardBaseAPIHandler):

    @authenticated
    async def delete(self, dashboard_urlname):

        current_user = await self.get_current_user()

        dashboard = self.db.query(Dashboard).filter(Dashboard.urlname==dashboard_urlname).one_or_none()

        if dashboard is None:
            raise HTTPError(404, "Dashboard does not exist.")

        if dashboard.user.name != current_user.name:
            raise HTTPError(403, 'This is not your dashboard: dashboard user {} does not match you ({})'.format(dashboard.user.name, current_user.name))

        # options = self.get_json_body()

        builders_store = BuildersStore.get_instance(self.settings['config'])

        builder = builders_store[dashboard]

        if builder.pending:
            raise HTTPError(400, 'Dashboard {} is currently building. Please try again once complete.'.format(dashboard_urlname))

        try:
        
            await self.maybe_delete_existing_server(dashboard.final_spawner, current_user)

            # TODO check delete went OK

            self.log.info("Deleting dashboard %s", dashboard_urlname)

            self.db.expire(dashboard)

            if dashboard in builders_store:
                del builders_store[dashboard]

            self.db.delete(dashboard)

            self.db.commit()

        except Exception as e:
            raise HTTPError(500, 'Error removing dashboard {}: {}'.format(dashboard_urlname, e))

        self.set_header('Content-Type', 'text/plain')
        self.set_status(202)


class UserSelfAPIHandler(DashboardBaseAPIHandler):

    @authenticated
    async def get(self):
        """Return the current user."""

        current_user = await self.get_current_user()

        self.set_status(200)
        self.finish(json.dumps(self.user_model(current_user)))



from jupyterhub.spawner import LocalProcessSpawner

from .variablemixin import VariableMixin, MetaVariableMixin


class VariableLocalProcessSpawner(LocalProcessSpawner, VariableMixin, metaclass=MetaVariableMixin):
    pass




from jupyterhub.handlers.pages import HomeHandler
from .base import DashboardBaseMixin

class OurHomeHandler(HomeHandler, DashboardBaseMixin):

    async def get(self, *args, **kwargs):

        current_user = await self.get_current_user()

        if not self.can_user_spawn(current_user):
            html = self.render_template(
                "homecds.html",
                base_url=self.settings['base_url'],
                current_user=current_user
            )
            return self.write(html)

        return await super().get(*args, **kwargs)
    


from kubespawner import KubeSpawner

from .variablemixin import VariableMixin, MetaVariableMixin, Command


class VariableKubeSpawner(KubeSpawner, VariableMixin, metaclass=MetaVariableMixin):

    default_presentation_cmd = Command(
        ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'],
        allow_none=False,
        help="""
        The command to run presentations through jhsingle_native_proxy, can be a string or list.
        Default is ['python3', '-m', 'jhsingle_native_proxy.main']
        Change to e.g. ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'] to ensure start hooks are
        run in the singleuser Docker images.
        """
    ).tag(config=True)
    
    def get_pvc_manifest(self):
        presentation_type = self._get_presentation_type()

        pvc = super().get_pvc_manifest()
        
        if presentation_type != '':
            pass
            #pvc.spec.data_source = {'kind': 'PersistentVolumeClaim', 'name': 'claim-dan'}
        
        return pvc
